// Copyright (C) 2006 JackTheDipper <gybrush@punk-ass-bitch.org>
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "configloader.h"

#include <fstream>
#include <iostream>
#include <sys/types.h>
#include <sys/stat.h>
//#include <errno.h>
#include <dirent.h> // not platform independent :(
#include <gtk/gtk.h>

/*
ConfigLoader::ConfigLoader()
{
//  engines = NULL;
}

ConfigLoader::~ConfigLoader()
{
  
}
*/

bool ConfigLoader::load_config(string filename, TreeHandler* m_refConfig, const Glib::RefPtr<Gnome::Glade::Xml>& m_refGlade, MainWindow* m_refWindow)
{
//todo: check use of strtol() !

  if(!ConfigLoader::check_file(filename)) // if file does not exists (or is no regular file)
    return 0;

  /*parse the file and get the DOM */
   xmlDoc* doc = xmlReadFile(filename.c_str(), NULL, 0);

  if(doc == NULL)
    return 0;
  
  if(strcmp((char*)xmlDocGetRootElement(doc)->name, "theme"))
  {
    cout << "Sorry, this file doesn't describe a real theme but a \"" << xmlDocGetRootElement(doc)->name << "\"!" << endl;
    return 0;
  }
    
  xmlNode* element = xmlDocGetRootElement(doc)->children;
  xmlChar* id;
  xmlChar* value;
  xmlChar* disabled;
  while( element != NULL )
  {

    if(element->type == XML_ELEMENT_NODE)
    {
      id = NULL;
      value = NULL;
      disabled = NULL;

      if(!strcmp((char*)element->name, "element"))
      {
        id = xmlGetProp(element, (xmlChar*)"id");
        value = xmlGetProp(element, (xmlChar*)"value");
        disabled = xmlGetProp(element, (xmlChar*)"disabled");
        
        if(id && value && strcmp((char*)id, "") && strcmp((char*)value, ""))
        {
          if(disabled && !strcasecmp((char*)disabled, "true"))
            m_refConfig->setOverride(strtol((char*)id, NULL, 10), false);
          else
            m_refConfig->setOverride(strtol((char*)id, NULL, 10), true);

          ConfigLoader::fill_config(m_refConfig, m_refGlade, m_refWindow, atoi((char*)id), (char*)value);
        }
        
        xmlFree(id);
        xmlFree(value);
        xmlFree(disabled);
      }
      else if(!strcmp((char*)element->name, "engine"))
      {
        id = xmlGetProp(element, (xmlChar*)"name");
        value = xmlGetProp(element, (xmlChar*)"category");
        disabled = xmlGetProp(element, (xmlChar*)"disabled");

        if(id && value && strcmp((char*)id, "") && strcmp((char*)value, ""))
        {
          // cout << "found engine " << id << " for category " << value << endl; // debugging ;-)
          
          if(disabled && !strcasecmp((char*)disabled, "true"))
            ConfigLoader::attach_engine((char*)id, (char*)value, m_refConfig, false);          
            // m_refConfig->setEngineOverride((char*)value, false);
          else
            ConfigLoader::attach_engine((char*)id, (char*)value, m_refConfig, true);
            // m_refConfig->setEngineOverride((char*)value, true);
          
          reload_enginetable_row(string((char*)value), m_refWindow);
          select_engine(string((char*)value), string((char*)id), m_refWindow);


          xmlNode* subelement = element->children;
          xmlChar* subname;
          xmlChar* subvalue;
          xmlChar* subdisabled;
          while( subelement != NULL )
          {
            subname = NULL;
            subvalue = NULL;
            subdisabled = NULL;
            
            if(subelement->type == XML_ELEMENT_NODE)
            {
              subname = xmlGetProp(subelement, (xmlChar*)"name");
              subvalue = xmlGetProp(subelement, (xmlChar*)"value");
              subdisabled = xmlGetProp(subelement, (xmlChar*)"disabled");
              
              if(subname && subvalue && strcmp((char*)subname, "") && strcmp((char*)subvalue, ""))
              {
                // cout << "-> found subelement " << subname << endl; // debugging ;-)

                //todo: set value only if 100% valid ;-)
                // (but theoretically it should be corrected automatically by the widgets themself)
                m_refConfig->setValue((char*)value, (char*)subname, (char*)subvalue); // value ^= category

                if(subdisabled && !strcasecmp((char*)subdisabled, "true"))
                  m_refConfig->setOverride((char*)value, (char*)subname, false);
                else
                  m_refConfig->setOverride((char*)value, (char*)subname, true);

              }
              
              xmlFree(subname);
              xmlFree(subvalue);
              xmlFree(subdisabled);
            }

            subelement = subelement->next;
          }
          
          
        }
        
        xmlFree(id);
        xmlFree(value);
        xmlFree(disabled);
      }

    }
    
    element = element->next;
  }

  return 1;

}

void ConfigLoader::fill_config(TreeHandler* m_refConfig, const Glib::RefPtr<Gnome::Glade::Xml>& m_refGlade, MainWindow* m_refWindow, int id, char* value)
{
//todo: safer implementation (atoi)

  char* type;
  ModWidget* widget = NULL;
  ModWidget* cbox = NULL;

  char id_string[11];
  snprintf (id_string, 10, "%i", id);
  
  m_refConfig->gotoElementById(id);
  type = m_refConfig->getElementType();

  if(!type)
  {
    cerr << "Cannot load element (error in database!)" << endl;
    return;
  }

  // cout << "found a " << type << endl; // debug ;-)

  bool valid_widget = true;

  //let the widgets itself and MainWindow correct wrong values and then save them
  if(!strcmp(type, "spin") || !strcmp(type, "iconsize"))
  {
    if(atoi(value) >= m_refConfig->getElementMin() && atoi(value) <= m_refConfig->getElementMax()) // not necessarily needed
    {
      char temp[11];
      snprintf (temp, 10, "%i", atoi(value));
      m_refConfig->setElementValue((char*)temp);

      m_refGlade->get_widget(string("spinbutton_").append(id_string), widget);
      widget->reload();
      if(!strcmp(type, "spin"))
        m_refWindow->on_changed_spinbutton(id, (SpinButton*)widget);
      else
        m_refWindow->on_changed_spinbutton_iconsize(id, (SpinButton*)widget);
    }
  }
  else if(!strcmp(type, "check_alone"))
  {
//todo: remove check_alone in favour of check and check_group
  }
  else if(!strcmp(type, "color"))
  {
    m_refConfig->setElementValue(value);
    m_refGlade->get_widget(string("colorbutton_").append(id_string), widget);
    widget->reload();
    m_refWindow->on_changed_colorbutton(id, (ColorButton*)widget);
  }
  else if(!strcmp(type, "font"))
  {
    m_refConfig->setElementValue(value);
    m_refGlade->get_widget(string("fontbutton_").append(id_string), widget);
    widget->reload();
    m_refWindow->on_changed_fontbutton(id, (FontButton*)widget);
  }
  else if(!strcmp(type, "image"))
  {
//todo: detect correct absolute path! and convert it or set an icon path!
    m_refConfig->setElementValue(value);
    m_refGlade->get_widget(string("filechooserbutton_").append(id_string), widget);
    widget->reload();
    m_refWindow->on_changed_filechooserbutton(id, (FileChooserButton*)widget, id_string);
  }
  else
  {
    cerr << "ID " << id << " has unknown type \"" << type << "\"!" << endl;
    valid_widget = false;
  }
   
  if(valid_widget)
  {
    cbox = m_refGlade->get_widget(string("checkbutton_").append(id_string), widget);
    cbox->reload();
  }

}


void ConfigLoader::reload_enginetable_row(string category, MainWindow* p_refWindow)
{
  int i = 0;
  while(p_refWindow->engine_name[i] != "")
  {
    if(p_refWindow->engine_name[i] == category)
    {
      p_refWindow->engine_cbox[i].reload();
      break;
    }
    i++;
  }
}

void ConfigLoader::select_engine(string category, string engine, MainWindow* p_refWindow)
{
  int i = 0;
  while(p_refWindow->engine_name[i] != "")
  {
    if(p_refWindow->engine_name[i] == category)
    {
      p_refWindow->engine_combo[i].set_active_text(engine);
      break;
    }
    i++;
  }
}


string ConfigLoader::xml_encode(string input)
{
  xmlChar *temp = xmlEncodeSpecialChars	(NULL, (xmlChar*)input.c_str());
  string copy = string((char*)temp);
  xmlFree(temp);
  
  return copy;
}


bool ConfigLoader::export_config(TreeHandler* m_refConfig, string filename, bool export_disabled_elements, bool export_comments)
{

  if(! (ConfigLoader::check_file(filename, true) || ConfigLoader::create_file(filename)) )
  {
    cerr << "cannot write to " << filename << " !" << endl;
    return 0;
  }
  // if a writable _file_ with that name exists or could be created

  ofstream* gtkrc = new ofstream( filename.c_str() );

  *gtkrc << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<theme>\n\n";

  m_refConfig->gotoFirstCategory();

  for(int i = 0; i < m_refConfig->getCategoryCount(); i++)
  {
    if(export_disabled_elements || m_refConfig->getCategoryOverride())
    {
      if(export_comments)
        *gtkrc << "  <!-- Category: " << ConfigLoader::xml_encode(m_refConfig->getCategoryName())
               << (m_refConfig->getCategoryShortDescription() ? string("(") + m_refConfig->getCategoryShortDescription() + string(")") : "")
               << " -->\n";



      m_refConfig->gotoCategoriesFirstElement();
      for(int j = 0; j < m_refConfig->getCategoryElementsCount(); j++)
      {
        if(export_disabled_elements || m_refConfig->getElementOverride())
        {
          if(export_comments)

            *gtkrc << "    <!-- Element: " << (m_refConfig->getElementLongName() ? \
                          ConfigLoader::xml_encode(m_refConfig->getElementLongName()) : \
                          ConfigLoader::xml_encode(m_refConfig->getElementName()) )
                   << (m_refConfig->getElementShortDescription() ? \
                          string("(") + ConfigLoader::xml_encode(m_refConfig->getElementShortDescription()) + string(")") : "")
                   << " -->\n";
                   
          *gtkrc << "  <element id =\"" << ConfigLoader::xml_encode(m_refConfig->getElementIdString()) << "\""
                 << " value=\"" << ConfigLoader::xml_encode(m_refConfig->getElementValue()) << "\""
                 << ((!m_refConfig->getElementOverride()) ? " disabled=\"true\"" : "")
                 << "/>\n";
        }
        m_refConfig->gotoCategoriesNextElement();
      }

      *gtkrc << "\n";
    }

    m_refConfig->gotoCategoriesEngine();
    char* engine = m_refConfig->getEngineName();
    if(engine != NULL && (export_disabled_elements || m_refConfig->getEngineOverride()) )
    {
//todo: check for more detailed comments (long names, blah blah)
      if(export_comments)
        *gtkrc << "    <!-- Engine: " << ConfigLoader::xml_encode(engine) << " -->\n";

      *gtkrc << "  <engine name=\"" << ConfigLoader::xml_encode(engine) << "\""
             << " category=\"" << ConfigLoader::xml_encode(m_refConfig->getCategoryName()) << "\""
             << ((!m_refConfig->getEngineOverride()) ? " disabled=\"true\"" : "")
             << ">\n";

      m_refConfig->gotoCategoryEnginesFirstParam();
      for(int j = 0; j < m_refConfig->getCategoryEnginePropertiesCount(); j++)
      {
        if(export_disabled_elements || m_refConfig->getParamOverride())
        {
          if(export_comments)
            *gtkrc << "      <!-- Param: " << (m_refConfig->getParamLongName() ? \
                                ConfigLoader::xml_encode(m_refConfig->getParamLongName()) : \
                                ConfigLoader::xml_encode(m_refConfig->getParamName()) )
                   << (m_refConfig->getParamShortDescription() ? \
                                string(" (") + ConfigLoader::xml_encode(m_refConfig->getParamShortDescription()) + string(")") : "")
                   << " -->\n";

          *gtkrc << "    <element name=\"" << ConfigLoader::xml_encode(m_refConfig->getParamName()) << "\""
                 << " value=\"" << ConfigLoader::xml_encode(m_refConfig->getParamValue()) << "\""
                 << ((!m_refConfig->getParamOverride()) ? " disabled=\"true\"" : "")
                 << "/>\n";
        }
        m_refConfig->gotoCategoryEnginesNextParam();
      }

      *gtkrc << "  </engine>\n\n\n";
    }



    
    m_refConfig->gotoNextCategory();
  }
  
  
  *gtkrc << "</theme>" << endl;

  delete gtkrc;
  
  return 1;
}


// returns true if at least one engine could be loaded!
bool ConfigLoader::load_engines(string directory, TreeHandler* m_refConfig)
{
  bool loaded_engine = false;
  struct dirent *entry;

  DIR *dir = opendir(directory.c_str());
  if(!dir)
    perror((string("Error in accessing ") + directory).c_str());
  else
  {
    while(entry = readdir(dir))
    {
      string name = string(entry->d_name);
      if(name.length() > 4)
      {
        if(name.substr(name.length()-4) == ".xml")
        {
          if(load_engine(directory + name))
            loaded_engine = true;
          else
            cout << "Couldn't load engine schema " << directory + name << endl;
        }
      }
    }
    closedir(dir);
  }

  

  return loaded_engine;
}


int ConfigLoader::load_engine(string filename)
{
  if(!ConfigLoader::check_file(filename)) // if file does not exists (or is no regular file)
    return 0;

  /*parse the file and get the DOM */
   xmlDoc* doc = xmlReadFile(filename.c_str(), NULL, 0);

  if(doc == NULL)
    return 0;
  
  xmlNode* root = xmlDocGetRootElement(doc);

  if(root == NULL)
    return 0;

  if(strcmp((char*)root->name, "engine"))
  {
    xmlFreeDoc(doc);
    return 0;
  }
  
  xmlChar* xmlchar_val = xmlGetProp(root, (xmlChar*)"module_name");

  if(xmlchar_val == NULL || !strcmp((char*)xmlchar_val, ""))
  {
    xmlFree(xmlchar_val);
    xmlFreeDoc(doc);
    return 0;
  }
  
  std::string name = std::string((char*)xmlchar_val);
  xmlFree(xmlchar_val);

  if(name == "") // just for being super clean ;-)
  {
    xmlFreeDoc(doc);
    return 0;
  }


  // check if elements have name, type, and default value defined, otherwise delete them off the tree!
  xmlNode* element = root->children;
  xmlNode* element_tmp = NULL;
  while( element != NULL )
  {
    element_tmp = NULL;
    
    if(element->type == XML_ELEMENT_NODE && !strcmp((char*)element->name, "element"))
    {
      xmlchar_val = xmlGetProp(element, (xmlChar*)"name");
      if(xmlchar_val == NULL || !strcmp((char*)xmlchar_val, ""))
      {
        element_tmp = element->next;
        xmlUnlinkNode(element);
        xmlFreeNode(element);
      }
      else
      {
        xmlFree(xmlchar_val);
        xmlchar_val = xmlGetProp(element, (xmlChar*)"type");
        if(xmlchar_val == NULL || !strcmp((char*)xmlchar_val, ""))
        {
          element_tmp = element->next;
          xmlUnlinkNode(element);
          xmlFreeNode(element);
        }
        else
        {
          xmlFree(xmlchar_val);
          xmlchar_val = xmlGetProp(element, (xmlChar*)"default");
          if(xmlchar_val == NULL || !strcmp((char*)xmlchar_val, ""))
          {
            element_tmp = element->next;
            xmlUnlinkNode(element);
            xmlFreeNode(element);
          }
          //else cout << "Valid Element!" << endl;
        }
      }
      xmlFree(xmlchar_val);
    }

    if(element_tmp)
      element = element_tmp;
    else
      element = element->next;
  }



  // save engine for later use
  Engine *engine;
  if(engines == NULL)
  {
    engines = new Engine;
    engine = engines;
  }
  else
  {
    engine = engines;
    while(engine->next != NULL)
    {
      if(engine->name == name)
      {
        xmlFreeDoc(doc);
        return 0;
      }
      engine = engine->next;
    }
    if(engine->name == name)
      return 0;

    engine->next = new Engine;
    engine = engine->next;
  }

//todo: add only engines with different names _AND_ lnames!

  engine->name = name;
  engine->doc = doc;
  engine->next = NULL;
  
  xmlChar* long_name = xmlGetProp(root, (xmlChar*)"long_name");
  engine->lname = (long_name != NULL) ? (char*)long_name : "";
  xmlFree(long_name);
}


void ConfigLoader::unload_engines() // deletes all engine objects, but, of course, doesn't delete insertions into the db xml tree! ;-)
{
  Engine *engine = engines;
  Engine *engine_next = NULL;

  while(engine != NULL)
  {
    engine_next = engine->next;
    xmlFreeDoc(engine->doc);
    delete engine;
    engine = engine_next;
  }

  engines = NULL;
}


void ConfigLoader::attach_engine(string engine_name, string category, TreeHandler* config, bool override)
{
  //cout << "!! attach engine " << engine_name << " to " << category << endl; // debugging, heh ;-)
  Engine *engine = engines;
  while(engine != NULL)
  {
    if(engine->name == engine_name)
    {
      config->attachEngine(xmlDocGetRootElement(engine->doc), (char*)category.c_str());
      if(override)
        config->setEngineOverride(category, true);
      break;
    }
    engine = engine->next;
  }
}


bool ConfigLoader::check_file(string filename, bool write_check)
{
  if(filename == "")
    return 0;

  struct stat config_stat;

  if(!stat( filename.c_str(), &config_stat) && config_stat.st_mode & S_IFREG) // if a regular file (no directory, etc)
  {
    fstream config( filename.c_str() , ios_base::in);

    if(config)
    {
      config.close();

      if(write_check)
      {
        config.open(filename.c_str(), ios_base::out | ios_base::app); // reopen for write
        if(!config)
          return 0;
        config.close();
      }

      return 1;
    }
  }

  return 0;
}


bool ConfigLoader::check_directory(string directory)
{
  struct stat config_stat;

  if(directory != "" && !stat( directory.c_str(), &config_stat) && (config_stat.st_mode & S_IFDIR))
    return 1;
  
  return 0;
}


bool ConfigLoader::create_file(string filename)
{
//  errno = 0;
  struct stat config_stat;
  
  if(filename != "" && !stat( filename.c_str(), &config_stat)) // if a file/directory or whatever with this filename exists
    return 0;
    
  ofstream config( filename.c_str() );
  if(config)
  {
    config.close();
    return 1;
  }
  
  return 0;
}


bool ConfigLoader::create_directory(string directory)
{
//  errno = 0; 
  struct stat config_stat;
  
  if(directory != "" && !stat( directory.c_str(), &config_stat)) // if a file/directory or whatever with this filename exists
    return 0;
    
  if(mkdir(directory.c_str(), (mode_t)00755) < 0)
    return 0;
    
  return 1;
}


bool ConfigLoader::remove_directory(string directory)
{
  struct stat config_stat;
  struct dirent *entry;

  if(ConfigLoader::check_directory(directory))
  {
    if(directory.substr(directory.length()-1,1) != "/")
      directory += "/";
  
    DIR *dir = opendir(directory.c_str());
    if(!dir)
      return 0;
    else
    {
    
      while(entry = readdir(dir))
      {
      
        string name = string(entry->d_name);

        if(name != "." && name != "..")
        {
          if(ConfigLoader::check_directory(directory + name))
          {
            if(!ConfigLoader::remove_directory(directory + name))
            {
              closedir(dir);
              return 0;
            }
          }
          else if(!stat( (directory + name).c_str(), &config_stat))
          {
            if(!ConfigLoader::remove_node(directory + name))
            {
              closedir(dir);
              return 0;
            }
          }
        }
        
      }
      
    }
    closedir(dir);

    if(!ConfigLoader::remove_node(directory))
      return 0;

    return 1;
  }
  else
    return 0;
}


bool ConfigLoader::remove_node(string path)
{
  struct stat config_stat;

//  if(!stat( path.c_str(), &config_stat) && !unlink(path.c_str())) // removes files/links only
  if(!stat( path.c_str(), &config_stat) && !remove(path.c_str())) // removes files/links and directories
    return 1;
  else
    return 0;
}


bool ConfigLoader::copy_directory(string src, string dst)
{
  struct stat config_stat;
  struct dirent *entry;

  if(src.substr(src.length()-1,1) != "/")
    src += "/";

  if(dst.substr(dst.length()-1,1) != "/")
    dst += "/";

  if(!ConfigLoader::create_directory(dst))
    return 0;


  if(ConfigLoader::check_directory(src))
  {

    DIR *dir = opendir(src.c_str());
    if(!dir)
      return 0;
    else
    {
    
      while(entry = readdir(dir))
      {
      
        string name = string(entry->d_name);

        if(name != "." && name != "..")
        {
          if(ConfigLoader::check_directory(src + name))
          {
            if(!ConfigLoader::copy_directory(src + name, dst + name))
            {
              closedir(dir);
              return 0;
            }
          }
          else if(!stat( (src + name).c_str(), &config_stat))
          {
            if(!ConfigLoader::copy_file(src + name, dst + name))
            {
              closedir(dir);
              return 0;
            }
          }
        }
        
      }
      
    }
    closedir(dir);

    return 1;
  }
  else
    return 0;
  
}


bool ConfigLoader::copy_file(string src, string dst)
{
  if(!ConfigLoader::check_file(src))
    return 0;
    
  if(! (ConfigLoader::check_file(dst, true) || ConfigLoader::create_file(dst)) )
    return 0;

  ifstream input(src.c_str());
  if(!input)
    return 0;
    
  fstream output(dst.c_str(), ios::out|ios::binary);
  if(!output)
    return 0;

  output << input.rdbuf();

  input.close();
  output.close();

  return 1;
}


bool ConfigLoader::check_include(string filename, string include, string searchpattern)
{
  if(!ConfigLoader::check_file(filename, true))
    return 0;
    
  fstream fi(filename.c_str(), ios_base::in);
  
  if(!fi)
    return 0;

  //scan
  string line;

  while ( getline ( fi, line ) )
  {
    //string::size_type i = line.find_first_not_of ( " \t\n\v" );

    if ( line == include || (searchpattern != "" && line.find(searchpattern, 0) != string::npos) )
    {
      fi.close();
      return 1;
    }
  }
  fi.close();

  return 0;
}


bool ConfigLoader::create_include(string filename, string include, string searchpattern)
{

  if(!ConfigLoader::check_file(filename, true))
    return 0;

  if(!ConfigLoader::check_include(filename, include, searchpattern))
  {
    ofstream gtkrc(filename.c_str(), ios_base::app);
    if(!gtkrc)
      return 0;

    gtkrc << include << "\n";
    gtkrc.close();
  }

}



void ConfigLoader::load_engine_params(EngineWindow* window, string engine_name, string category, TreeHandler* config)
{
  bool skipParam;

  if(engine_name != "" && engines != NULL)
  {
    Engine *engine = engines;
    
    while(engine != NULL)
    {
      if(engine_name == engine->name)
        break;
      engine = engine->next;
    }
    
    if(engine != NULL && engine_name == engine->name)
    {
      xmlNode* root = xmlDocGetRootElement(engine->doc);
      xmlNode* element = root->children;
      int y = 0;
      
      while(element != NULL)
      {
    	  if(element->type == XML_ELEMENT_NODE && !strcmp((char*)element->name,"option"))
    	  {
    	    skipParam = false;

/*    	    
          xmlChar *c_name  = xmlGetProp(element, (xmlChar*)"name"),
                  *c_value = xmlGetProp(element, (xmlChar*)"value"),
                  *c_type  = xmlGetProp(element, (xmlChar*)"type"),
//                  *c_lname = xmlGetProp(element, (xmlChar*)"long_name"),
//                  *c_sdesc = xmlGetProp(element, (xmlChar*)"short_description"),
//                  *c_ldesc = xmlGetProp(element, (xmlChar*)"long_description"),
                  //*c_min = xmlGetProp(element, (xmlChar*)"min"),
                  //*c_max = xmlGetProp(element, (xmlChar*)"max"),
                  //*c_yes = xmlGetProp(element, (xmlChar*)"yes"),
                  //*c_no = xmlGetProp(element, (xmlChar*)"no"),
                  *c_disabled = xmlGetProp(element, (xmlChar*)"disabled");

    	    string w_name  = c_name  ? (char*)c_name  : "",
    	           w_value = c_value ? (char*)c_value : "",
    	           w_type  = c_type  ? (char*)c_type  : "";
//    	           w_lname = c_lname ? (char*)c_lname : "",
//    	           w_sdesc = c_sdesc ? (char*)c_sdesc : "",
//    	           w_ldesc = c_ldesc ? (char*)c_ldesc : "";
    	           //w_min = c_min ? (char*)c_min : "",
    	           //w_max = c_max ? (char*)c_max : "",
    	           //w_yes = c_yes ? (char*)c_yes : "",
    	           //w_no = c_no ? (char*)c_no : "";

    	    
    	    xmlFree(c_name);
    	    xmlFree(c_value);
    	    xmlFree(c_type);
//    	    xmlFree(c_lname);
//    	    xmlFree(c_sdesc);
//    	    xmlFree(c_ldesc);
    	    //xmlFree(c_min);
    	    //xmlFree(c_max);
    	    //xmlFree(c_yes);
    	    //xmlFree(c_no);
    	    xmlFree(c_disabled);
*/    	    
    	    const char* dummy = NULL;
     	    string w_name     = (dummy = get_property(element, "name"))             ? dummy : "",
    	           w_value    = (dummy = get_property(element, "value"))            ? dummy : "",
    	           w_default  = (dummy = get_property(element, "default"))          ? dummy : "",
    	           w_type     = (dummy = get_property(element, "type"))             ? dummy : "";
    	    string w_lname    = (dummy = get_content(element,  "long_name"))        ? dummy : "",
    	           w_sdesc    = (dummy = get_content(element,  "description"))      ? dummy : "",
    	           w_ldesc    = (dummy = get_content(element,  "comment"))          ? dummy : "";
          bool   w_disabled = ((dummy = get_property(element, "disabled")) && !strcasecmp(dummy, "yes")) ? true : false;    	    


          // create new Parameter (for saving widgets in order to be able to delete them later)
          Param *param;
          if(params == NULL)
          {
            params = new Param();
            param = params;
          }
          else
          {
            param = params;
            while(param->next != NULL)
            {
              if(param->name == w_name)
              {
                skipParam = true;
              }
              param = param->next;
            }

            if(param->name == w_name)
            {
              skipParam = true;
            }
    
            if(skipParam)
            {
              element = element->next;
              break; // goto next parameter
            }
            param->next = new Param();
            param = param->next;
          }

          param->name = w_name;
          param->next = NULL;



          // attention: recycling of skipParam *g*
          skipParam = false;



          // create widgets of that new Parameter and attach them to the output table


/*
          if(w_type == "spin")
          {
            SpinButton *widget = new SpinButton();
            widget->init(category, w_name, config);
            //widget->reload(); // already run by init() !
            param->widget = (Gtk::Widget*)widget;
            widget->signal_value_changed().connect( sigc::bind( sigc::bind( sigc::mem_fun(*window, &EngineWindow::on_changed_spinbutton), widget) , w_name) );

          }
*/
          if(w_type == "color")
          {
            ColorButton *widget = new ColorButton();
            widget->init(category, w_name, config);
            param->widget = (Gtk::Widget*)widget;
            widget->signal_color_set().connect( sigc::bind( sigc::bind( sigc::mem_fun(*window, &EngineWindow::on_changed_colorbutton), widget) , w_name) );
          }
          else if(w_type == "boolean")
          {
            CheckButton *widget = new CheckButton();
            widget->init(category, w_name, config);
            param->widget = (Gtk::Widget*)widget;
            widget->signal_toggled().connect( sigc::bind( sigc::bind( sigc::mem_fun(*window, &EngineWindow::on_changed_checkbutton), widget) , w_name) );
          }
          else if(w_type == "integer")
          {
            Slider *widget = new Slider();
            widget->init(category, w_name, config);
            param->widget = (Gtk::Widget*)widget;
            widget->signal_value_changed().connect( sigc::bind( sigc::bind( sigc::mem_fun(*window, &EngineWindow::on_changed_slider), widget) , w_name) );
          }
          else if(w_type == "real")
          {
            Slider *widget = new Slider();
            widget->init(category, w_name, config);
            param->widget = (Gtk::Widget*)widget;
            widget->signal_value_changed().connect( sigc::bind( sigc::bind( sigc::mem_fun(*window, &EngineWindow::on_changed_slider), widget) , w_name) );
          }
          else if(w_type == "enumeration")
          {
            ComboBox *widget = new ComboBox();
            if(w_ldesc != "")
            {
              param->widget_ebox = new Gtk::EventBox();
              param->widget_ebox->add(*widget);
            }

            
            for(xmlNode* option = element->children; option != NULL; option = option->next)
            {
              if(option->type == XML_ELEMENT_NODE && !strcmp((char*)option->name, "enumeration"))
              {
                //option_value = xmlGetProp(option, (xmlChar*)"value");
                //option_name  = xmlNodeGetContent(option);
                string option_value = (dummy = get_property(option, "value")) ? dummy : "",
                       option_name  = (dummy = get_content(option, "label")) ? dummy : "";
                
//                if(option_value && option_name && strcmp((char*)option_value, "") && strcmp((char*)option_name, ""))
                if(option_value != "" && option_name != "")
                {
                  widget->append_text((char*)option_value.c_str(), (char*)option_name.c_str());
                }
//                xmlFree(option_value);
//                xmlFree(option_name);
//                option_value = NULL;
//                option_name  = NULL;
              }
            }

            widget->init(category, w_name, config);
            param->widget = (Gtk::Widget*)widget;
            widget->signal_changed().connect( sigc::bind( sigc::bind( sigc::mem_fun(*window, &EngineWindow::on_changed_combobox), widget) , w_name) );
          }
          else
          {
            //todo: delete param and its widgets!
            skipParam = true;
          }


          if(!skipParam)
          {
            char* config_label = NULL;
          
            param->box   = new Gtk::HBox();
//            param->label = new Gtk::Label(w_lname != "" ? w_lname : param->name);
            param->label = new Gtk::EventBox();
            param->label->add_label(w_lname != "" ? w_lname : param->name);
//            param->desc  = new Gtk::Label(w_sdesc != "" ? w_sdesc : (Glib::ustring("Configuration of ") +
//                                         (w_lname != "" ? w_lname : param->name)).c_str(), 0.0, 0.0);
            param->desc = new Gtk::EventBox();
//            param->desc->add_label(w_sdesc != "" ? w_sdesc : (Glib::ustring("Configuration of ") +
//                                  (w_lname != "" ? w_lname : param->name)).c_str(), 0.0, 0.0);

            param->desc->add_label(
                  w_sdesc != "" ? w_sdesc : config_label = g_strdup_printf(_("Configuration of %s"),
                        (w_lname != "" ? w_lname.c_str() : param->name.c_str() )),
                  0.0, 0.0);

            if(config_label)
              free(config_label);
            
            ((Gtk::Label*)(param->desc->get_child()))->set_line_wrap(true);
            param->line  = new Gtk::HSeparator();
            param->line->set_size_request(-1, 20);
            
            param->cbox  = new CheckButton();
            param->cbox->init(category, w_name, config, false);
            param->cbox->signal_toggled().connect( sigc::bind( sigc::bind( sigc::bind( sigc::mem_fun(*window, &EngineWindow::on_checkbox_toggled), param->widget),  param->cbox), w_name) );
            
            if(w_ldesc != "")
            {
              window->tooltips.set_tip(*param->label,  w_ldesc);
              window->tooltips.set_tip(*param->desc,   w_ldesc);
              if(param->widget_ebox)
                window->tooltips.set_tip(*param->widget_ebox, w_ldesc);
              else
                window->tooltips.set_tip(*param->widget, w_ldesc);
            }
            
            window->table.resize(y+2, 2); // rows, coloumns
            window->table.attach(*param->desc,  0, 2, y,   y+1,  Gtk::FILL |  Gtk::EXPAND, ~Gtk::FILL & ~Gtk::EXPAND);
            window->table.attach(*param->label, 0, 1, y+1, y+2, ~Gtk::FILL & ~Gtk::EXPAND, ~Gtk::FILL & ~Gtk::EXPAND);
            window->table.attach(*param->box,   1, 2, y+1, y+2,  Gtk::FILL |  Gtk::EXPAND, ~Gtk::FILL & ~Gtk::EXPAND);
            window->table.attach(*param->line,  0, 2, y+2, y+3,  Gtk::FILL |  Gtk::EXPAND, ~Gtk::FILL & ~Gtk::EXPAND);
            
            param->box->pack_start(*param->cbox,   Gtk::PACK_SHRINK);
            if(param->widget_ebox)
              param->box->pack_start(*param->widget_ebox, Gtk::PACK_EXPAND_WIDGET);
            else
              param->box->pack_start(*param->widget, Gtk::PACK_EXPAND_WIDGET);
              
          }


          y = y+3;

    	  }
    	  element = element->next;
      }
      
    }
  }
}



void ConfigLoader::unload_engine_params()
{
  Param *param = params;
  Param *param_next;
  
  while(param != NULL)
  {
    param_next = param->next;
    if(param->box)
      delete param->box;
    if(param->label)
      delete param->label;
    if(param->desc)
      delete param->desc;
    if(param->line)
      delete param->line;
    if(param->cbox)
      delete param->cbox;
    if(param->widget)
      delete param->widget;
    if(param->widget_ebox)
      delete param->widget_ebox;
    delete param;
    param = param_next;
  }

  params = NULL;
  // cout << "unloaded engine" << endl; // debugging ;-)
}

void ConfigLoader::fill_with_engines(string category, ComboBox* box, TreeHandler* p_refConfig)
{
  if(box != NULL)
  {
    //box->append_text("<none>");
    if(engines != NULL)
    {
      bool first_entry = 1;
      Engine *engine = engines;
      while(engine != NULL)
      {
        box->append_text(engine->name, (engine->lname != "") ? engine->lname : engine->name);
        if(first_entry)
        {
          box->set_active_text(engine->name);
          first_entry = false;
        }
        engine = engine->next;
      }

      if(box->get_active_text() != "")
        ConfigLoader::attach_engine(box->get_active_text(), category, p_refConfig, false);
    }
  }
}


ConfigLoader::Param::Param()
{
  this->box    = NULL;
  this->label  = NULL;
  this->desc   = NULL;
  this->line   = NULL;
  this->cbox   = NULL;
  this->widget = NULL;
  this->widget_ebox = NULL;
  this->next   = NULL;
}




/***********************************************************/
//todo: remove these ugly doubled methods and create statics

const char* ConfigLoader::get_property(xmlNode* node, char* property)
{
  if(node)
  {    
    xmlChar* value = xmlGetProp(node, (xmlChar*)property);

    if(value != NULL && strcmp((char*)value, ""))
    {
      string temp = string((char*)value);
      xmlFree(value);
      return temp.c_str();
    }

    xmlFree(value);
  }

  return NULL;
}

const char* ConfigLoader::get_content(xmlNode* node, const char* content)
{
  if(node && content)
  {
    xmlNode *child = NULL;
    char *loc = NULL;
    string locale;
//todo: only use locale for indeed translatable tags
    if( content \
        && (    !strcmp(content, "long_name") \
             || !strcmp(content, "description") \
             || !strcmp(content, "comment") \
             || !strcmp(content, "label")
           )
        && (loc = setlocale(LC_ALL, NULL)) && strcmp(loc, "en")
      )
    {
      const char* output = NULL;
      locale = loc;

      if(output = get_lang(get_node(node->children, content), content, locale.c_str()))
        return output;

      if(locale.find("@", 0) != string::npos)
        if(output = get_lang(get_node(node->children, content), content, locale.substr(0, locale.find("@", 0)).c_str() ))
          return output;

      if(locale.find(".", 0) != string::npos)
        if(output = get_lang(get_node(node->children, content), content, locale.substr(0, locale.find(".", 0)).c_str() ))
          return output;

      if(locale.find("_", 0) != string::npos)
        if(output = get_lang(get_node(node->children, content), content, locale.substr(0, locale.find("_", 0)).c_str() ))
          return output;
    }
    
    return get_lang(get_node(node->children, content), content, NULL, true);
  }

  return NULL;
}


const char* ConfigLoader::get_lang(xmlNode* node, const char* content, const char* locale, bool acceptSpace)
{
  xmlNode* child = node;
  xmlChar *lang = NULL;

  while(child != NULL && content != NULL)
  {
    lang = xmlNodeGetLang(child);
    if(acceptSpace || (lang && !strcmp((char*)lang, locale)))
    {
      //if(!acceptSpace)cout << "got content with " << locale << " for " << child->name << endl;
      //else cout << "got default content for " << child->name << endl;
      
      xmlChar* value = xmlNodeGetContent(child);
      if(value != NULL && strcmp((char*)value, ""))
      {
        string temp = string((char*)value);
        xmlFree(value);
        xmlFree(lang);
        return temp.c_str();
      }
      xmlFree(value);
    }
    //else cout << "got no content with " << locale << " for " << child->name << endl;
    xmlFree(lang);
    child = get_node(child->next, content);
  }

  return NULL;
}



xmlNode* ConfigLoader::get_node(xmlNode* ref_node, string name)
{
  if(ref_node)
  {
    xmlNode* node = ref_node;
  
    while(node != NULL && !(node->type == XML_ELEMENT_NODE && !strcmp((char*)node->name, name.c_str())))
      node = node->next;
   
    if(node != NULL && node->type == XML_ELEMENT_NODE && !strcmp((char*)node->name, name.c_str()))
      return node;
  }

  return NULL;
}






