// Copyright (C) 2006 JackTheDipper <gybrush@punk-ass-bitch.org>
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "gtkrcexporter.h"
#include "configloader.h"

#include <fstream>
using namespace std;


GtkrcExporter::GtkrcExporter(TreeHandler* config)
{
  m_refConfig = config;
}


GtkrcExporter::~GtkrcExporter()
{

}


/*** interface methods ***/

void GtkrcExporter::export_to(string filename, string image_path)
{

  if(! (ConfigLoader::check_file(filename, true) || ConfigLoader::create_file(filename)) )
  {
    //cerr << "cannot write to " << filename << " !" << endl;
    throw IoException(IoException::WriteError);
  }


//  string image_path = filename.substr(0, filename.rfind("/") - 1).append("images/");

  string icons;
  ofstream* gtkrc = new ofstream(filename.c_str());

  *gtkrc << "pixmap_path \"" << image_path << "\"" << endl << endl;


  m_refConfig->gotoFirstCategory();

  for(int i = 0; i < m_refConfig->getCategoryCount(); i++)
  {
    if(m_refConfig->getCategoryOverride())
    {
//      if(!strcmp(m_refConfig->getCategoryName(), "default")) // default-style
//        *gtkrc << "style \"gnome-color-chooser-default\"\n{\n";
//      else if(strcmp(m_refConfig->getCategoryName(), "_TOPLEVEL_")) // and if not a toplevel command
//        *gtkrc << "style \"gnome-color-chooser-" << m_refConfig->getCategoryName() << "\" = \"gnome-color-chooser-default\"\n{\n";

      if(strcmp(m_refConfig->getCategoryName(), "_TOPLEVEL_")) // if not a toplevel command
        *gtkrc << "style \"gnome-color-chooser-" << m_refConfig->getCategoryName() << "\"\n{\n";

      icons = "";

      m_refConfig->gotoCategoriesEngine();
      if(m_refConfig->getEngineName() != NULL && m_refConfig->getEngineOverride())
      {
        *gtkrc << "  engine \"" << m_refConfig->getEngineName() << "\"\n  {\n";

        m_refConfig->gotoCategoryEnginesFirstParam();
        if(m_refConfig->getParamName() != NULL)
        {
            while(m_refConfig->getParamName() != NULL)
            {
              if(m_refConfig->getParamOverride())
              {
                *gtkrc << "    " << m_refConfig->getParamName()
                       << " = " << (use_quotes(m_refConfig->getParamType()) ? "\"" : "")
                       << m_refConfig->getParamValue()
                       << (use_quotes(m_refConfig->getParamType()) ? "\"" : "")
                       << "\n";
              }
              m_refConfig->gotoCategoryEnginesNextParam();
            }

        }

        *gtkrc << "  }\n\n";
      }

      m_refConfig->gotoCategoriesFirstElement();
      for(int j = 0; j < m_refConfig->getCategoryElementsCount(); j++)
      {
        if(m_refConfig->getElementOverride())
        {
          if(!strcmp(m_refConfig->getElementType(), "iconsize"))
          {
            if(icons == "")
              icons = "gtk-icon-sizes = \"";
            icons.append((string)m_refConfig->getElementName() + "=" + (string)m_refConfig->getElementValue() + ":");
          }
          else
          {
            *gtkrc << (strcmp(m_refConfig->getCategoryName(), "_TOPLEVEL_") ? "  " : "")
                   << m_refConfig->getElementName()
                   << " = " << (use_quotes(m_refConfig->getElementType()) ? "\"" : "")
                   << m_refConfig->getElementValue()
                   << (use_quotes(m_refConfig->getElementType()) ? "\"" : "")
                   << "\n";
          }

        }
        m_refConfig->gotoCategoriesNextElement();
      }

      *gtkrc << (icons != "" ? icons.append("\"") : "")
             << (!strcmp(m_refConfig->getCategoryName(), "_TOPLEVEL_") ? "\n" : "}\n");

      if(strcmp(m_refConfig->getCategoryName(), "_TOPLEVEL_"))
      {
        m_refConfig->gotoCategoriesFirstClass();
        for(int j = 0; j < m_refConfig->getCategoryClassesCount(); j++)
        {
          *gtkrc << m_refConfig->getClassType() << " \"" << m_refConfig->getClassName() << "\" style \"gnome-color-chooser-" << m_refConfig->getCategoryName() << "\"\n";
          m_refConfig->gotoCategoriesNextClass();
        }
      }

      *gtkrc << "\n";
    }
    m_refConfig->gotoNextCategory();
  }

  delete gtkrc;



}



/*** private methods ***/

bool GtkrcExporter::use_quotes(char* type)
{
//todo: add engine param option: quote !!! which overrides this setting for most types!
  if(!strcmp(type, "color"))
    return true;
  else if(!strcmp(type, "spin"))
    return false;
  else if(!strcmp(type, "font"))
    return true;
  else if(!strcmp(type, "iconsize"))
    return false; // is externally quoted
  else if(!strcmp(type, "check"))
    return false;
  else if(!strcmp(type, "slider"))
    return false;
  else if(!strcmp(type, "select"))
    return false;
  else if(!strcmp(type, "check_alone"))
    return false;

// additional engine option types
  else if(!strcmp(type, "boolean"))
    return false;
  else if(!strcmp(type, "enumeration"))
    return false;
  else if(!strcmp(type, "integer"))
    return false;
  else if(!strcmp(type, "real"))
    return false;

  return true;
}


