// Copyright (C) 2006 JackTheDipper <gybrush@punk-ass-bitch.org>
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "treehandler.h"
#include <iostream>


TreeHandler::TreeHandler(char* file)
{
  this->filename = file;
  doc = NULL;
  root_node = NULL;
  cat_cur = NULL;
  el_cur = NULL;
  class_cur = NULL;
  param_cur = NULL;
  eng_cur = NULL;

  /*
   * this initialize the libxml2 library and check potential ABI mismatches
   * between the version it was compiled for and the actual shared
   * library used.
   */
  LIBXML_TEST_VERSION

  /*parse the file and get the DOM */
  doc = xmlReadFile(filename, NULL, 0);

  if (doc == NULL)
    exit(1);
    //std::cerr << "error: could not parse file " << file << std::endl;
    
  /*Get the root element node */
  root_node = xmlDocGetRootElement(doc);

//  xmlSaveFormatFileEnc("-", doc, "UTF-8", 1);

/*
  this->version = 0;
  xmlChar* name = (xmlChar*)"version";
  xmlChar* xmlchar_version = xmlGetProp(doc->children, name);
  if(xmlchar_version != NULL)
    this->version = atoi((char*)xmlchar_version);

  xmlFree(xmlchar_version);  
*/
}

TreeHandler::~TreeHandler()
{
  /*free the document */
  xmlFreeDoc(doc);

  /*
   *Free the global variables that may
   *have been allocated by the parser.
   */
  xmlCleanupParser();
}

void TreeHandler::reset_to_defaults()
{
//  doc = NULL;
  root_node = NULL;
  cat_cur = NULL;
  el_cur = NULL;
  class_cur = NULL;
  param_cur = NULL;
  eng_cur = NULL;

  xmlDoc* new_doc = NULL;
  /*parse the file and get the DOM */
  new_doc = xmlReadFile(filename, NULL, 0);

  if (new_doc == NULL)
    std::cerr << "error: could not parse file " << filename << std::endl;
  else
  {
    doc = new_doc;
    /*Get the root element node */
    root_node = xmlDocGetRootElement(doc);

/*
  this->version = 0;
  xmlChar* name = (xmlChar*)"version";
  xmlChar* xmlchar_version = xmlGetProp(doc->children, name);
  if(xmlchar_version != NULL)
    this->version = atoi((char*)xmlchar_version);

  xmlFree(xmlchar_version);
*/
  }
}

/*
int TreeHandler::getConfigVersion()
{
  return this->version;
}

void TreeHandler::setConfigVersion(int version)
{
  this->version = version;

  char temp[11];
  xmlChar* name = (xmlChar*)"version";
  snprintf (temp, 10, "%i", version);
  xmlSetProp(doc->children, name, (xmlChar*)temp);
  //xmlSaveFormatFileEnc("-", doc, "UTF-8", 1); // debugging ;-)
}
*/

void TreeHandler::addCategory(xmlNode* category)
{
  if(category != NULL)
  {
    category = xmlDocCopyNode(category, doc, 1); // 1 means recursively (with all children)
    xmlAddChild(root_node, category);
    //xmlSaveFormatFileEnc("-", doc, "UTF-8", 1); // debugging ;-)
    cat_cur = category;
  }
}


void TreeHandler::addElement(xmlNode* element)
{
  if(cat_cur != NULL && element != NULL)
  {
    element = xmlDocCopyNode(element, doc, 2); // 2 means with all properties and namespaces, without children
    xmlAddChild(cat_cur, element);
    //xmlSaveFormatFileEnc("-", doc, "UTF-8", 1); // debugging ;-)
    el_cur = element;
  }
}


void TreeHandler::addClass(xmlNode* cl)
{
  if(cat_cur != NULL && cl != NULL)
  {
    cl = xmlDocCopyNode(cl, doc, 2);
    xmlAddChild(cat_cur, cl);
    //xmlSaveFormatFileEnc("-", doc, "UTF-8", 1); // debugging ;-)
    class_cur = cl;
  }
}

void TreeHandler::attachEngine(xmlNode* engine, char* cat_name)
{
  if(engine != NULL && cat_name != NULL)
  {
    xmlNode* eng_prev = get_engine(cat_name);
    if(eng_prev)
      this->unattachEngine(cat_name);

    xmlNode* category = get_category(cat_name);
    
    if(category != NULL)
    {
      engine = xmlDocCopyNode(engine, doc, 1); //1 means recursively (with all children)
      xmlAddChild(category, engine);
//      el_cur = engine;
//      xmlSetProp(el_cur, (xmlChar*)"category", (xmlChar*)cat_name);
//      xmlSaveFormatFileEnc("-", doc, "UTF-8", 1); // debugging ;-)
    }
  }
}


void TreeHandler::unattachEngine(string category)
{
  if(category != "")
  {
    xmlNode* engine = get_engine(category);
    if(engine)
    {
      xmlUnlinkNode(engine);
      xmlFreeNode(engine);
      eng_cur = NULL; // just for being save
    }
  }
}


xmlNode* TreeHandler::getCategoryNode()
{
  return cat_cur;
}


xmlNode* TreeHandler::getElementNode()
{
  return el_cur;
}

xmlNode* TreeHandler::getClassNode()
{
  return class_cur;
}


int TreeHandler::getCategoryCount()
{
  int i = 0;
	
	for(xmlNode* category = get_first_category(); // count all categories (first category and its siblings)
		category != NULL;
		category = get_next_category(category))
	{
      i++;
	}

  return i;
}


void TreeHandler::gotoFirstCategory()
{
	cat_cur = get_first_category();
}

void TreeHandler::gotoNextCategory() //goto next sibling (node on same level)
{
  cat_cur = get_next_category(cat_cur);
}

bool TreeHandler::getCategoryOverride()
{
	for(xmlNode* element = cat_cur->children;
		element != NULL;
		element = element->next)
	{
	  if(element->type == XML_ELEMENT_NODE && (!strcmp((char*)element->name,"element") || !strcmp((char*)element->name, "engine")))
	  {
      if( (r_return = get_property(element, "override")) && !strcmp(r_return, "yes") )
        return true;
    }
	}

  return false;
}

char* TreeHandler::getCategoryName()
{
  return get_property(m_out_cat_name, cat_cur, "name");
}


char* TreeHandler::getCategoryShortDescription()
{
  return this->get_property(cat_cur, "short_description");
}


char* TreeHandler::getCategoryLongName()
{
  return this->get_property(cat_cur, "long_name");
}


int TreeHandler::getCategoryElementsCount()
{
  int i = 0;
	
	for(xmlNode* element = cat_cur->children;
		element != NULL;
		element = element->next)
	{
	  if(element->type == XML_ELEMENT_NODE && !strcmp((char*)element->name,"element"))
      i++;
	}

  return i;
}

int TreeHandler::getCategoryClassesCount()
{
  int i = 0;
	
	for(xmlNode* element = cat_cur->children;
		element != NULL;
		element = element->next)
	{
	  if(element->type == XML_ELEMENT_NODE && !strcmp((char*)element->name,"class"))
      i++;
	}

  return i;
}

//

void TreeHandler::gotoCategoriesFirstElement()
{
  el_cur = NULL;
	xmlNode* element = cat_cur->children;
  while(element != NULL)
	{
	  if(element->type == XML_ELEMENT_NODE && !strcmp((char*)element->name,"element"))
	  {
      el_cur = element;
      break;
    }
    element = element->next;
	}
}


void TreeHandler::gotoCategoriesNextElement()
{
  if(el_cur != NULL)
  {
    xmlNode* el = NULL;
    el = el_cur->next;
    while(el != NULL && !(el->type == XML_ELEMENT_NODE && !strcmp((char*)el->name,"element")))
      el = el->next;
      
    if(el != NULL && el->type == XML_ELEMENT_NODE && !strcmp((char*)el->name,"element"))
      el_cur = el;
    else
      el_cur = NULL;
  }
}


void TreeHandler::gotoCategoriesEngine()
{
  eng_cur = get_engine(cat_cur);
}


void TreeHandler::gotoCategoryEnginesParamByName(char* name)
{
  param_cur = get_param(get_engine(cat_cur), name);
}


void TreeHandler::gotoCategoriesEngine(string category)
{
  eng_cur = get_engine(category);
}


char* TreeHandler::getEngineName()
{
  return this->get_property(m_out_eng_name, eng_cur, "module_name");
}


char* TreeHandler::getEngineName(string category)
{
  return this->get_property(m_out_eng_name, get_engine(category), "module_name");
}


void TreeHandler::gotoCategoryEnginesFirstParam()
{
  param_cur = get_first_param(get_engine(cat_cur));
}


void TreeHandler::gotoCategoryEnginesNextParam()
{
  param_cur = get_next_param(param_cur);
}


int TreeHandler::getCategoryEnginePropertiesCount()
{
  int i = 0;

  if(eng_cur != NULL)
  {
	  for(xmlNode* element = eng_cur->children;
	  	element != NULL;
	  	element = element->next)
	  {
	    if(element->type == XML_ELEMENT_NODE && !strcmp((char*)element->name,"option"))
        i++;
	  }
  }
  
  return i;
}


char* TreeHandler::getParamName()
{
  return this->get_property(m_out_param_name, param_cur, "name");
}


char* TreeHandler::getParamValue()
{
  if(char* value = this->get_property(m_out_param_val, param_cur, "value"))
    return value;
  else
    return this->get_property(m_out_param_val, param_cur, "default");
}


const char* TreeHandler::getParamLongName()
{
  return this->get_content(m_out_param_lname, param_cur, "long_name");
}

const char* TreeHandler::getParamShortDescription()
{
  return this->get_content(m_out_param_sdesc, param_cur, "description");
}


bool TreeHandler::getParamOverride()
{
  return ( (r_return = get_property(param_cur, "override")) && !strcmp(r_return, "yes")) ? true : false;
}

char* TreeHandler::getParamType()
{
  return this->get_property(m_out_param_type, param_cur, "type");
}

/*
void TreeHandler::setParamOverride(string category, string id, bool override)
{
  this->setOverride(get_param(get_engine(category), id), override);
}
*/

bool TreeHandler::getEngineOverride()
{
  return ( (r_return = get_property(eng_cur, "override")) && !strcmp(r_return, "yes")) ? true : false;
}

bool TreeHandler::getEngineOverride(string category)
{
  return ( (r_return = get_property(get_engine(category), "override")) && !strcmp(r_return, "yes")) ? true : false;
}

void TreeHandler::setEngineOverride(string category, bool override)
{
  //cout << "set engine override for category " << category << " to " << override <<  endl; // debugging ;-)
  this->setOverride(get_engine(category), override);
  //xmlSaveFormatFileEnc("-", doc, "UTF-8", 1); // debugging ;-)
}


void TreeHandler::gotoFirstElement()
{
  this->gotoFirstCategory();

  while(cat_cur != NULL)
  {
    this->gotoCategoriesFirstElement();
    if(el_cur != NULL)
    {
      break;
    }
    this->gotoNextCategory();
  }

}


void TreeHandler::gotoNextElement()
{
  this->gotoCategoriesNextElement();

  while(cat_cur != NULL)
  {
    if(el_cur != NULL)
      break;
    this->gotoNextCategory();
    if(cat_cur != NULL)
      this->gotoCategoriesFirstElement();
  }
 
}


void TreeHandler::gotoElementById(int id)
{
  el_cur = getElementById(id);
}


int TreeHandler::getElementCount() // overwrites current category and current element !!
{
  this->gotoFirstElement();
  int i = 0;
  while(el_cur != NULL)
  {
    this->gotoNextElement();
    i++;
  }
  return i;
}


bool TreeHandler::getElementOverride()
{
  return ( (r_return = get_property(el_cur, "override")) && !strcmp(r_return, "yes")) ? true : false;
}

char* TreeHandler::getElementIdString()
{
  return this->get_property(m_out_el_id, el_cur, "id");
}

char* TreeHandler::getElementName()
{
  return this->get_property(m_out_el_name, el_cur, "name");
}


char* TreeHandler::getElementLongName()
{
  return this->get_property(m_out_el_lname, el_cur, "long_name");
}


char* TreeHandler::getElementValue()
{
  return this->get_property(m_out_el_val, el_cur, "value");
}

char* TreeHandler::getElementType()
{
  return this->get_property(m_out_el_type, el_cur, "type");
}


int TreeHandler::getElementMin()
{
  return ( (r_return = get_property(el_cur, "min")) && strcmp(r_return, "")) ? atoi(r_return) : 0;
}


int TreeHandler::getElementMax()
{
  return ( (r_return = get_property(el_cur, "max")) && strcmp(r_return, "")) ? atoi(r_return) : 255;
}


char* TreeHandler::getElementShortDescription()
{
  return this->get_property(el_cur, "short_description");
}


void TreeHandler::setElementValue(char* value)
{
  if(el_cur != NULL)
  {
    xmlSetProp(el_cur, (xmlChar*)"value", (xmlChar*)value);
    //this->setOverride(el_cur, true);
    //xmlSaveFormatFileEnc("-", doc, "UTF-8", 1); // debugging ;-)
  }
  else
    std::cerr << "error: could not set value" << std::endl;
}



//

void TreeHandler::gotoCategoriesFirstClass()
{
  class_cur = NULL;
	xmlNode* element = cat_cur->children;
  while(element != NULL)
	{
	  if(element->type == XML_ELEMENT_NODE && !strcmp((char*)element->name,"class"))
	  {
      class_cur = element;
      break;
    }
    element = element->next;
	}
}

void TreeHandler::gotoCategoriesNextClass()
{
  if(class_cur != NULL)
  {
      xmlNode* el = NULL;
      el = class_cur->next;
      while(el != NULL && !(el->type == XML_ELEMENT_NODE && !strcmp((char*)el->name,"class")))
        el = el->next;
      if(el != NULL && el->type == XML_ELEMENT_NODE && !strcmp((char*)el->name,"class"))
        class_cur = el;
      else
        class_cur = NULL;
  }
}

char* TreeHandler::getClassName()
{
  return this->get_property(m_out_class_name, class_cur, "name");
}

char* TreeHandler::getClassType()
{
  return this->get_property(m_out_class_type, class_cur, "type");
}

char* TreeHandler::getValue(int id)
{
  return this->get_property(m_out_param_val, getElementById(id), "value");
}


char* TreeHandler::getValue(string category, string id)
{
  // cout << "getValue " << category << " " << id << endl; // debugging ;-)
//todo: testrun with unknown category and/or id !

  if(char* value = get_property(m_out_param_val, get_param(get_engine(category), id), "value"))
  {
    return value;cout << value << endl;}
  else
    return get_property(m_out_param_val, get_param(get_engine(category), id), "default");
}


void TreeHandler::setValue(string category, string id, char* value)
{
  // cout << "setValue " << category << " " << id << " " << value << endl; // debugging ;-)

  if(xmlNode* node = get_param(get_engine(category), id))
    xmlSetProp(node, (xmlChar*)"value", (xmlChar*)value);
}


bool TreeHandler::getOverride(int id) // element
{
  return ( (r_return = get_property(getElementById(id), "override")) && !strcmp(r_return, "yes")) ? true : false;
}

bool TreeHandler::getOverride(string category, string id) // engine param
{
  return ( (r_return = get_property(get_param(get_engine(category), id), "override")) && !strcmp(r_return, "yes")) ? true : false;
}

void TreeHandler::setValue(int id, char* value)
{
  xmlNode* el = getElementById(id);
  if(el != NULL)
  {
    xmlSetProp(el, (xmlChar*)"value", (xmlChar*)value);
    //this->setOverride(el, true);
    //xmlSaveFormatFileEnc("-", doc, "UTF-8", 1); // debugging ;-)
  }
  else
    std::cout << "error: element " << id << " not found!" << std::endl;
}

void TreeHandler::setOverride(int id, bool override)
{
  this->setOverride(getElementById(id), override);
}

void TreeHandler::setOverride(xmlNode* element, bool override)
{
  if(element != NULL)
  {
    xmlSetProp(element, (xmlChar*)"override", (xmlChar*)(override ? "yes" : "no"));
    //xmlSaveFormatFileEnc("-", doc, "UTF-8", 1); // debugging ;-)
  }
}

void TreeHandler::setOverride(string category, string id, bool override)
{
  this->setOverride(get_param(get_engine(category), id), override);
}


int TreeHandler::getMin(int id)
{
  return ( (r_return = get_property(getElementById(id), "min")) && strcmp(r_return, "")) ? atoi(r_return) : 0;
}


int TreeHandler::getMax(int id)
{
  return ( (r_return = get_property(getElementById(id), "max")) && strcmp(r_return, "")) ? atoi(r_return) : 255;
}


int TreeHandler::getMin(string category, string id)
{
  // cout << "getMin " << category << " " << id << endl; // debugging ;-)
//todo: testrun with unknown category and/or id !

  return ( (r_return = get_content(get_param(get_engine(category), id), "minimum")) && strcmp(r_return, "")) ? atoi(r_return) : 0;
}


int TreeHandler::getMax(string category, string id)
{
  // cout << "getMax " << category << " " << id << endl; // debugging ;-)
//todo: check what happens with invalid integer in atoi function !!
//todo: testrun with unknown category and/or id !

  return ( (r_return = get_content(get_param(get_engine(category), id), "maximum")) && strcmp(r_return, "")) ? atoi(r_return) : 255;
}


char* TreeHandler::getChecked(string category, string id)
{
// currently hardcoded due to the lack of support from gtkrc
  return "TRUE";
  //return (char*)get_content(m_out_param_checked, get_param(get_engine(category), id), "checked");
}

char* TreeHandler::getUnchecked(string category, string id)
{
// currently hardcoded due to the lack of support from gtkrc
  return "FALSE";
  //return (char*)get_content(m_out_param_unchecked, get_param(get_engine(category), id), "unchecked");
}


int TreeHandler::getFractionalDigits(string category, string id)
{
  char* type = getType(category, id);
  if(type && !strcmp(type, "integer"))
    return 0;
  else // "real" with default precision 1
    return ( (r_return = get_content(get_param(get_engine(category), id), "precision")) && strcmp(r_return, "")) ? atoi(r_return) : 1;
}


char* TreeHandler::getType(string category, string id)
{
  return this->get_property(get_param(get_engine(category), id), "type");
}


xmlNode* TreeHandler::getElementById(int id)
{
	xmlNode* category = get_first_category();
  xmlNode* element = NULL;
  char* element_id = NULL;
  

  while(category != NULL)
  {
    element = category->children;
    while(element != NULL)
    {
      if(element->type == XML_ELEMENT_NODE && !strcmp((char*)element->name,"element") && ((element_id = get_property(element, "id")) && atoi(element_id) == id)) // if found an element -> compare id
      {
        return element;
      }
      element = element->next;
    }
    category = get_next_category(category);
  }

  return NULL;
}


void TreeHandler::save()
{
  xmlSaveFormatFileEnc(this->filename, doc, "UTF-8", 1);
}

/******************************************************************************/


char* TreeHandler::get_property(xmlNode* node, char* property)
{
  if(node)
  {    
    xmlChar* value = xmlGetProp(node, (xmlChar*)property);

    if(value != NULL && strcmp((char*)value, ""))
    {
      string temp = string((char*)value);
      xmlFree(value);
      return (char*)temp.c_str();
    }

    xmlFree(value);
  }

  return NULL;
}

char* TreeHandler::get_property(string &m_out, xmlNode* node, char* property)
{
  if(node)
  {
    xmlChar* value = xmlGetProp(node, (xmlChar*)property);

    if(value != NULL && strcmp((char*)value, ""))
    {
      m_out.assign((char*)value);
      xmlFree(value);
      return (char*)m_out.c_str();
    }

    xmlFree(value);
  }

  return NULL;
}

const char* TreeHandler::get_content(xmlNode* node, char* content)
{
  if(node && content)
  {
    xmlNode *child = get_node(node->children, content);
    if(child == NULL)
      return NULL;
    
    xmlChar* value = xmlNodeGetContent(child);
    if(value != NULL && strcmp((char*)value, ""))
    {
      string temp = string((char*)value);
      xmlFree(value);
      return temp.c_str();
    }
    
    xmlFree(value);
  }

  return NULL;
}

const char* TreeHandler::get_content(string &m_out, xmlNode* node, char* content)
{
  const char* value = get_content(node, content);
  
  if(value != NULL && strcmp(value, ""))
  {
    m_out.assign(value);
    return m_out.c_str();
  }
  
  return NULL;
}


xmlNode* TreeHandler::get_category(string name)
{
  if(name != "")
  {
    xmlNode* category = doc->children->children;
    char* cat_name = NULL;
    
    while(category != NULL && !((cat_name = get_property(category, "name")) && !strcmp(cat_name, name.c_str())) )
      category = get_next_category(category);

    if(category != NULL && cat_name != NULL && !strcmp(cat_name, name.c_str()))
      return category;
  }
  
  return NULL;
}

xmlNode* TreeHandler::get_first_category()
{
  return get_node(doc->children->children, "category");
}

xmlNode* TreeHandler::get_next_category(xmlNode* category)
{
  return get_node(category->next, "category");
}

xmlNode* TreeHandler::get_engine(string category_name)
{
  return get_engine(get_category(category_name));
}

xmlNode* TreeHandler::get_engine(xmlNode* category)
{
  if(category)
  {
  	xmlNode* eng = category->children;
    while(eng != NULL && !(eng->type == XML_ELEMENT_NODE && !strcmp((char*)eng->name, "engine")))
      eng = eng->next;

    if(eng != NULL && eng->type == XML_ELEMENT_NODE && !strcmp((char*)eng->name, "engine"))
      return eng;
  }

  return NULL;
}

xmlNode* TreeHandler::get_param(xmlNode* engine, string name)
{
  if(engine && name != "")
  {
    xmlNode* param = engine->children;
    char* param_name = NULL;
    
    while(param != NULL && !((param_name = get_property(param, "name")) && !strcmp(param_name, name.c_str())) )
      param = get_next_param(param);

    if(param != NULL && param_name != NULL && !strcmp(param_name, name.c_str()))
      return param;
  }
  
  return NULL;
}

xmlNode* TreeHandler::get_first_param(xmlNode* engine)
{
  return get_node(engine->children, "option");
}

xmlNode* TreeHandler::get_next_param(xmlNode* param)
{
  return get_node(param->next, "option");
}

xmlNode* TreeHandler::get_node(xmlNode* ref_node, string name)
{
  if(ref_node)
  {
    xmlNode* node = ref_node;
  
    while(node != NULL && !(node->type == XML_ELEMENT_NODE && !strcmp((char*)node->name, name.c_str())))
      node = node->next;

    if(node != NULL && node->type == XML_ELEMENT_NODE && !strcmp((char*)node->name, name.c_str()))
      return node;
  }

  return NULL;
}


//todo@all: use strtol or something similar instead of atoi!!


