// Copyright (C) 2006 JackTheDipper <gybrush@punk-ass-bitch.org>
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software 
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef _TREEHANDLER_H
#define _TREEHANDLER_H

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <string>

using namespace std;

class TreeHandler
{
  public:
    TreeHandler(char* file);
    ~TreeHandler();
    void save();
    void reset_to_defaults();
    
//    int getConfigVersion();
//    void setConfigVersion(int version);
    
    void addCategory(xmlNode* category);
    void addElement(xmlNode* element);
    void addClass(xmlNode* cl);
    void attachEngine(xmlNode* engine, char* cat_name);
    void unattachEngine(string category);

    int getCategoryCount();
    xmlNode* getCategoryNode();
    xmlNode* getElementNode();
    xmlNode* getClassNode();

    void gotoFirstCategory();
    void gotoNextCategory();
    bool getCategoryOverride();
    char* getCategoryName();
    char* getCategoryLongName();
    char* getCategoryShortDescription();
    int getCategoryElementsCount();
    int getCategoryClassesCount();
    void gotoCategoriesFirstElement();
    void gotoCategoriesNextElement();

    void gotoCategoriesEngine();
    void gotoCategoriesEngine(string category);
    void gotoCategoryEnginesFirstParam();
    void gotoCategoryEnginesNextParam();
    void gotoCategoryEnginesParamByName(char* name);
    int getCategoryEnginePropertiesCount();
    char* getEngineName();
    char* getEngineName(string category);
    char* getParamName();
    char* getParamValue();
    char* getParamType();
    const char* getParamLongName();
    const char* getParamShortDescription();

    bool getEngineOverride();
    bool getEngineOverride(string category);
    void setEngineOverride(string category, bool override);
    bool getParamOverride();
    //void setParamOverride(string category, string id, bool override);

    void gotoFirstElement();
    void gotoNextElement();
    void gotoElementById(int id);
    int getElementCount(); // overwrites current category and current element !!
    bool getElementOverride();
    char* getElementIdString();
    char* getElementName();
    char* getElementLongName();
    char* getElementValue();
    char* getElementType();
    int   getElementMin();
    int   getElementMax();
    char* getElementShortDescription();
    
    void  setElementValue(char* value);
        
    void gotoCategoriesFirstClass();
    void gotoCategoriesNextClass();
    char* getClassName();
    char* getClassType();

    char* getValue(int id);
    void  setValue(int id, char* value);
    char* getValue(string category, string id);
    void  setValue(string category, string id, char* value);

    
    bool  getOverride(int id); // element
    bool  getOverride(string category, string id); // engine param

    void  setOverride(int id, bool override); // element
    void  setOverride(string category, string id, bool override); // engine param
    //char* getPath(int id);
    //void  setPath(int id, char* value);

    /* get-only params */
    int getMin(int id);
    int getMax(int id);
    int getMin(string category, string id);
    int getMax(string category, string id);
    char* getChecked(string category, string id);
    char* getUnchecked(string category, string id);
    int getFractionalDigits(string category, string id);
    char* getType(string category, string id);


  private:
    char* filename;
//    int version;
    xmlDoc *doc;
    xmlNode *root_node;
    xmlNode *cat_cur;
    xmlNode *el_cur;
    xmlNode *class_cur;
    xmlNode *param_cur;
    xmlNode *eng_cur;
    string m_out_cat_name;
    string m_out_cat_lname;
    string m_out_cat_sdesc;
    string m_out_el_id;
    string m_out_el_name;
    string m_out_el_val;
    string m_out_el_lname;
    string m_out_el_type;
    string m_out_el_sdesc;
    string m_out_class_name;
    string m_out_class_type;
    string m_out_eng_name;
    string m_out_param_name;
    string m_out_param_val;
    string m_out_param_lname;
    string m_out_param_sdesc;
    string m_out_param_checked;
    string m_out_param_unchecked;
    string m_out_param_type;
    const char *r_return;

    void setMaxCatId(int max);
    void setMaxElementId(int max);
    xmlNode* getElementById(int id);
    void setOverride(xmlNode* element, bool override);

    char* get_property(xmlNode* node, char* property);
    char* get_property(string &m_out, xmlNode* node, char* property);
    const char* get_content(xmlNode* node, char* content);
    const char* get_content(string &m_out, xmlNode* node, char* content);
    xmlNode* get_category(string name);
    xmlNode* get_first_category();
    xmlNode* get_next_category(xmlNode* category);
    xmlNode* get_engine(string category_name);
    xmlNode* get_engine(xmlNode* category);
    xmlNode* get_param(xmlNode* engine, string name);
    xmlNode* get_first_param(xmlNode* engine);
    xmlNode* get_next_param(xmlNode* param);
    xmlNode* get_node(xmlNode* ref_node, string name);
};

#endif // _TREEHANDLER_H
